/**********************************************************************/
/*                                                                    */
/*   Licensed Materials - Property of IBM.                            */
/*   IBM XL Fortran Advanced Edition V10.1 for Linux (5724-M17)       */
/*   Copyright IBM Corp. 2003, 2006.                                  */
/*   All Rights Reserved.                                             */
/*   US Government Users Restricted Rights -                          */
/*   Use, duplication or disclosure restricted by                     */
/*   GSA ADP Schedule Contract with IBM Corp.                         */
/*                                                                    */
/**********************************************************************/
/*
            NOTICE TO USERS OF THE SOURCE CODE EXAMPLES

 THE SOURCE CODE EXAMPLES PROVIDED BY IBM ARE ONLY INTENDED TO ASSIST IN THE
 DEVELOPMENT OF A WORKING SOFTWARE PROGRAM.  THE SOURCE CODE EXAMPLES DO NOT
 FUNCTION AS WRITTEN:  ADDITIONAL CODE IS REQUIRED.  IN ADDITION, THE SOURCE
 CODE EXAMPLES MAY NOT COMPILE AND/OR BIND SUCCESSFULLY AS WRITTEN.

 INTERNATIONAL BUSINESS MACHINES CORPORATION PROVIDES THE SOURCE CODE
 EXAMPLES, BOTH INDIVIDUALLY AND AS ONE OR MORE GROUPS, "AS IS" WITHOUT
 WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT
 LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE
 OF THE SOURCE CODE EXAMPLES, BOTH INDIVIDUALLY AND AS ONE OR MORE GROUPS,
 IS WITH YOU.  SHOULD ANY PART OF THE SOURCE CODE EXAMPLES PROVE
 DEFECTIVE, YOU (AND NOT IBM OR AN AUTHORIZED RISC System/6000* WORKSTATION
 DEALER) ASSUME THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR
 CORRECTION.

 IBM does not warrant that the contents of the source code examples, whether
 individually or as one or more groups, will meet your requirements or that
 the source code examples are error-free.

 IBM may make improvements and/or changes in the source code examples at
 any time.

 Changes may be made periodically to the information in the source code
 examples; these changes may be reported, for the sample exception handler
 included herein, in new editions of the examples.

 References in the source code examples to IBM products, programs, or
 services do not imply that IBM intends to make these available in all
 countries in which IBM operates.  Any reference to an IBM licensed
 program in the source code examples is not intended to state or imply
 that only IBM's licensed program may be used.  Any functionally equivalent
 program may be used.

 * RISC System/6000 is a trademark of International Business Machines
   Corporation.
*/

#if defined(__linux__) || defined(__APPLE__)
#define SIGNAME SIGFPE
#else
#define SIGNAME SIGTRAP
#endif

/*
 * COMPONENT_NAME: SIGNAME handler for the FLTTRAP compiler option
 *
 * FUNCTIONS: fhandler_
 *
 * ORIGINS: 27
 *
 */

/*
 * This code demonstrates a SIGNAME handler for use with the
 * 'flttrap' compiler option.  The handler includes the
 * ability to identify the point an exception occurred and
 * to continue execution following an exception, possibly
 * supplying a value as the result of the failing instruction.
 *
 * The name of the sample handler is 'fhandler_', with a trailing
 * underscore to enable use with the Fortran 'extname' compiler
 * option.
 *
 * This source code should be compiled using 'cc -c flttrap_handler.c'.
 */

#if defined(__linux__) || defined(__APPLE__)
/*
 * The flttrap compiler option will generate SIGFPE signals when
 * floating point exceptions occur.  It does so by setting the
 * the floating point enabled exception (FEX) bit in the floating
 * point status and control register (FPSCR).  The OS then traps
 * if the floating point exception (FX) summary bit in the FPSCR
 * is set.
 */
#else /* AIX */
/*
 * The flttrap compiler option will generate SIGTRAP signals when
 * floating point exceptions occur.  It does so by setting the
 * record bit on all floating point instructions, and then
 * trapping if condition register bit 5 is set (that is, if
 * the floating point enabled exception (FEX) bit is set in
 * the floating point status and control register).
 */
#endif

/*
 * To use this SIGNAME handler, the user code must be compiled
 * with the flttrap compiler option, must enable the desired
 * floating point exceptions (possibly with the flttrap 'enable'
 * suboption), and must install the 'fhandler_' routine as the
 * SIGNAME signal handler (possibly with the 'sigtrap' compiler
 * option).
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#if defined(__linux__) || defined(__APPLE__)
#include <sys/ucontext.h>

/* Mask for all FPSCR Exception bits */
#define FP_ALL_XCP      ((unsigned int) 0xFFF80700)

#else /* AIX */

#include <fptrap.h>
#include <fpxcp.h>

/* The specific trap instruction used by the flttrap option is
 * TRAP R15=R15.  This is the machine code for that instruction. */
#define FLTTRAPINST (0x7c8f7808)

/* The following variables record the location of the failing
 * operation, the kind of operation, and the floating point
 * registers found in a failing instruction.  Note that the
 * valid registers depend on the instruction type. */
static unsigned int *fpe_loc;
static char *opcode;
static int frt_reg, fra_reg, frb_reg, frc_reg;

/* Mask value to check for floating point exceptions. */
#define TST_MASK (FP_INVALID|FP_OVERFLOW|FP_UNDERFLOW| \
                  FP_DIV_BY_ZERO|FP_INEXACT)

#endif

/* The following table maps instruction bit patterns to the name
 * of a floating point instruction.  This table is referenced
 * using bits 26-30 of a floating point instruction. */
static char *op_table[32] =
    { "fcmp", "?", "?", "?", "?", "?", "?", "?",
      "?", "?", "?", "?", "frsp", "?", "?", "?",
      "?", "?", "fd", "?", "fs", "fa", "?", "?",
      "?", "fm", "?", "?", "fms", "fma", "fnms", "fnma" };

/*
 * NAME: fhandler_
 *
 * FUNCTION: sample SIGNAME exception handler
 *
 * NOTES: Customize this code by printing additional debugging
 *        information and defining exception results.
 */
#if defined(__linux__) || defined(__APPLE__)
void fhandler_(signal, info, context)
    int signal;
    siginfo_t *info;
    void *context;
{
#if defined(__APPLE__) && defined(__64BIT__)
    ucontext64_t *contextData = (ucontext64_t *) context;
#else
    ucontext_t *contextData = (ucontext_t *) context;
#endif

    /* The following variables record the location of the failing
     * operation, the kind of operation, and the floating point
     * registers found in a failing instruction.  Note that the
     * valid registers depend on the instruction type. */
    unsigned int *instr;
    char *opcode;
    int frt_reg, fra_reg, frb_reg, frc_reg;

    /* The 'fpscr' variable will be used to access the Floating
     * Point Status and Control Register. */
    unsigned int *fpscr;

    /* The 'fpr' variable will be used to access the floating-
     * point registers. */
    double *fpr;

#if defined(__linux__) && defined(__64BIT__)
    /* Unfortunately, the context provided by the OS does not
     * contain the proper FPSCR value, so we can't do much
     * with the fpscr variable here, and won't be able to
     * process more than one exceptions.
     */
    fpscr = ((unsigned int *)
            &(contextData->uc_mcontext.regs->gpr[PT_FPSCR])) + 1;
    instr = (unsigned int *) contextData->uc_mcontext.regs->nip;
    fpr = (double *) &(contextData->uc_mcontext.regs->gpr[PT_FPR0]);
#elif defined(__linux__) && !defined(__64BIT__)
    fpscr = ((unsigned int *)
            &(contextData->uc_mcontext.uc_regs->fpregs.fpscr)) + 1;
    instr = (unsigned int *) contextData->uc_mcontext.uc_regs->gregs[PT_NIP];
    fpr = (double *) contextData->uc_mcontext.uc_regs->fpregs.fpregs;
#elif defined(__APPLE__)
#if defined(__64BIT__)
    fpscr = (unsigned int *) &(contextData->uc_mcontext64->fs.fpscr);
    instr = (unsigned int *) contextData->uc_mcontext64->ss.srr0;
    fpr = (double *) contextData->uc_mcontext64->fs.fpregs;
#else
    fpscr = (unsigned int *) &(contextData->uc_mcontext->fs.fpscr);
    instr = (unsigned int *) contextData->uc_mcontext->ss.srr0;
    fpr = (double *) contextData->uc_mcontext->fs.fpregs;
#endif
#endif

    /* Check that the trap is of the type used for the flttrap option */
    if (signal != SIGFPE) {
        fprintf(stderr, "Unknown signal %d received.\n", signal);
        exit(42);
    }

    /* option and that a floating-point exception has occurred. */
    if (info->si_code != FPE_FLTDIV && info->si_code != FPE_FLTOVF &&
        info->si_code != FPE_FLTUND && info->si_code != FPE_FLTRES &&
        info->si_code != FPE_FLTINV) {
        fprintf(stderr, "SIGFPE without floating point exception\n");
        exit(42);
    }

    /* Validate the floating point instruction causing the exception */
    if (*instr >> 26 != 63) {
        fprintf(stderr, "SIGFPE handler failed to find exception point\n");
        /* Note that, since the exception may have occurred in a
         * subroutine that was not compiled with the flttrap option,
         * it may be desirable to simply ignore the exception by
         * clearing the exception bits and returning. */
        exit(43);
    }

    /* Decode the instruction to identify the kind of operation
     * and the source and result registers. */
    opcode = op_table[(*instr >> 1) & 0x1f];
    frt_reg = (*instr >> 21) & 0x1f;
    fra_reg = (*instr >> 16) & 0x1f;
    frb_reg = (*instr >> 11) & 0x1f;
    frc_reg = (*instr >>  6) & 0x1f;

    /* Examine the floating point status and control register for
     * enabled exceptions.  Customize each case below. */
    switch (info->si_code) {
    case FPE_FLTINV:
        fprintf(stderr,
                "FP invalid operation, operation '%s', location %0*lx\n",
                opcode, (int) (2*sizeof(instr)), (long) instr);
        /* Consider an invalid operation a fatal error.  Note that
         * by examining other bits in the status and control register
         * we could identify the specific invalid operation which
         * occurred (for example, zero divided by zero).  Using the
         * kind of operation, we could examine the source operands.
         * Note that any result register has not been modified. */
        exit(44);
    case FPE_FLTOVF:
        fprintf(stderr,
                "FP overflow, operation '%s', location %0*lx\n",
                opcode, (int) (2*sizeof(instr)), (long) instr);
        /* Note that the result register in an overflow contains a
         * correctly rounded normalized number, but 1536 has been
         * subtracted from the exponent.
         * Set the result of any overflow to zero. */
        fpr[frt_reg] = 0.0;
        break;
    case FPE_FLTUND:
        fprintf(stderr,
                "FP underflow, operation '%s', location %0*lx\n",
                opcode, (int) (2*sizeof(instr)), (long) instr);
        /* Note that the result register in an underflow contains a
         * correctly rounded normalized number, but 1536 has been
         * added to the exponent.
         * Set the result of any underflow to zero. */
        fpr[frt_reg] = 0.0;
        break;
    case FPE_FLTDIV:
        fprintf(stderr,
                "FP division by zero, operation '%s', location %0*lx\n",
                opcode, (int) (2*sizeof(instr)), (long) instr);
        /* Print the source operands for the division; the divide
         * instruction uses FRA and FRB.  Note that the result
         * register has not been modified by the divide. */
        fprintf(stderr,
                "    Division source operands: %f / %f\n",
                fpr[fra_reg], fpr[frb_reg]);
        /* Set the result of any division by zero to zero. */
        fpr[frt_reg] = 0.0;
        break;
    case FPE_FLTRES:
        fprintf(stderr,
                "FP inexact, operation '%s', location %0*lx\n",
                opcode, (int) (2*sizeof(instr)), (long) instr);
        /* No action, just ignore this. */
        break;
    }

    /* Reset the exception bits because they are sticky. */
    *fpscr &= ~FP_ALL_XCP;

    /* Continue execution with the instruction following the trap.
     * Don't try this one at home kids! */
#if defined(__linux__) && defined(__64BIT__)
    contextData->uc_mcontext.regs->nip += 4;
#elif defined(__linux__) && !defined(__64BIT__)
    contextData->uc_mcontext.uc_regs->gregs[PT_NIP] += 4;
#elif defined(__APPLE__)
#if defined(__64BIT__)
    contextData->uc_mcontext64->ss.srr0 += 4;
#else
    contextData->uc_mcontext->ss.srr0 += 4;
#endif
#endif
}

#else /* AIX */

void fhandler_(sig,code,scp)
    int sig, code;
    struct sigcontext *scp;
{
    fptrap_t fpstat;
    int result_reg;

    fpstat = scp->sc_jmpbuf.jmp_context.fpscr;

    /* Check that the trap is of the type used for the flttrap
     * option and that the floating point status and control
     * register indicates that an exception has occurred. */
    if (*((int *) scp->sc_jmpbuf.jmp_context.iar) != FLTTRAPINST ||
            !(fpstat & TST_MASK)) {
        /* This must be a trap caused by an integer division by
         * zero or a subscript out of range. */
        fputs("SIGTRAP without floating point exception\n",stderr);
        exit(42);
    }

    /* Find the floating point instruction causing the exception and
     * decode it.  find_inst sets the static variables which indicate
     * the instruction location, kind, and registers. */
    if (find_instr(scp->sc_jmpbuf.jmp_context.iar)) {
        fputs("SIGTRAP handler failed to find exception point\n",stderr);
        /* Note that, since the exception may have occurred in a
         * subroutine that was not compiled with the flttrap option,
         * it may be desirable to simply ignore the exception by
         * clearing the exception bits and returning. */
        exit(43);
    }

    /* Examine the floating point status and control register for
     * enabled exceptions.  Customize each case below. */
    if ((fpstat & (TRP_INVALID|FP_INVALID)) ==
            (TRP_INVALID|FP_INVALID)) {
        fprintf(stderr,
            "FP invalid operation, operation '%s', location %0*lx\n",
            opcode, (int) (2*sizeof(fpe_loc)), (long) fpe_loc);
        /* Consider an invalid operation a fatal error.  Note that
         * by examining other bits in the status and control register
         * we could identify the specific invalid operation which
         * occurred (for example, zero divided by zero).  Using the
         * kind of operation, we could examine the source operands.
         * Note that any result register has not been modified. */
        exit(44);
    }
    if ((fpstat & (TRP_OVERFLOW|FP_OVERFLOW)) ==
            (TRP_OVERFLOW|FP_OVERFLOW)) {
        fprintf(stderr,"FP overflow, operation '%s', location %0*lx\n",
            opcode, (int) (2*sizeof(fpe_loc)), (long) fpe_loc);
        /* Note that the result register in an overflow contains a
         * correctly rounded normalized number, but 1536 has been
         * subtracted from the exponent.
         * Set the result of any overflow to zero. */
        scp->sc_jmpbuf.jmp_context.fpr[frt_reg] = 0.0;
    }
    if ((fpstat & (TRP_UNDERFLOW|FP_UNDERFLOW)) ==
            (TRP_UNDERFLOW|FP_UNDERFLOW)) {
        fprintf(stderr,"FP underflow, operation '%s', location %0*lx\n",
            opcode, (int) (2*sizeof(fpe_loc)), (long) fpe_loc);
        /* Note that the result register in an underflow contains a
         * correctly rounded normalized number, but 1536 has been
         * added to the exponent.
         * Set the result of any underflow to zero. */
        scp->sc_jmpbuf.jmp_context.fpr[frt_reg] = 0.0;
    }
    if ((fpstat & (TRP_DIV_BY_ZERO|FP_DIV_BY_ZERO)) ==
            (TRP_DIV_BY_ZERO|FP_DIV_BY_ZERO)) {
        fprintf(stderr,
            "FP division by zero, operation '%s', location %0*lx\n",
            opcode, (int) (2*sizeof(fpe_loc)), (long) fpe_loc);
        /* Print the source operands for the division; the divide
         * instruction uses FRA and FRB.  Note that the result
         * register has not been modified by the divide. */
        fprintf(stderr,"    Division source operands: %f / %f\n",
            scp->sc_jmpbuf.jmp_context.fpr[fra_reg],
            scp->sc_jmpbuf.jmp_context.fpr[frb_reg]);
        /* Set the result of any division by zero to zero. */
        scp->sc_jmpbuf.jmp_context.fpr[frt_reg] = 0.0;
    }
    if ((fpstat & (TRP_INEXACT|FP_INEXACT)) ==
            (TRP_INEXACT|FP_INEXACT)) {
        fprintf(stderr,"FP inexact, operation '%s', location %0*lx\n",
            opcode, (int) (2*sizeof(fpe_loc)), (long) fpe_loc);
        /* No action, just ignore this. */
    }

    /* Reset the exception bits because they are sticky. */
    scp->sc_jmpbuf.jmp_context.fpscr &= ~FP_ALL_XCP;

    /* signal(SIGTRAP,fhandler_); */

    /* Continue execution with the instruction following the trap.
     * Don't try this one at home kids! */
    scp->sc_jmpbuf.jmp_context.iar += 4;
}


/* Find and decode the floating point instruction causing the
 * exception.  Return 1 if not found, else zero. */
static int find_instr(trap_loc)
    unsigned int *trap_loc;
{
    /* Search backward in the instruction stream starting from
     * trap_loc, looking for a floating point instruction (bits
     * 0-5 equal decimal 63).  The first such instruction found
     * will be assumed to be the failing operation.
     * Note that a linear backward search assumes that there is
     * no branching code separating the trap instruction from
     * the failing floating point operation.  This will always
     * be true with the current implementation of the flttrap
     * option (in fact, in the current implementation the
     * failing operation will always be the second last
     * instruction before the trap point), except in the case
     * of subroutine calls causing an exception.
     * For safety we limit the search length. */
    int i = 0;
    while ((*(--trap_loc) >> 26 != 63) &&
           (++i < 10));
    if (*trap_loc >> 26 != 63) return(1); /* no float op found */
    /* Check that the operation found has the record bit set. */
    if (!(*trap_loc & 1)) return(1); /* record bit not set */
    /* Check to see if the instruction found was a move register.
     * This instruction is produced after calls to external
     * routines to see if they returned with any exception bits
     * set.  Any such external routine must be a library routine
     * or in user code that was not compiled with flttrap. */
    if (((*trap_loc >> 1) & 0x3ff) == 72) return(1); /* fmr found */
    /* Decode the instruction to identify the kind of operation
     * and the source and result registers. */
    fpe_loc = trap_loc;
    opcode = op_table[(*trap_loc >> 1) & 0x1f];
    frt_reg = (*trap_loc >> 21) & 0x1f;
    fra_reg = (*trap_loc >> 16) & 0x1f;
    frb_reg = (*trap_loc >> 11) & 0x1f;
    frc_reg = (*trap_loc >> 6) & 0x1f;
    return(0);
}

#endif
