!***********************************************************************
!*   
!*   Licensed Materials - Property of IBM.
!*   IBM XL Fortran Advanced Edition V10.1 for Linux (5724-M17)
!*   Copyright IBM Corp. 2003, 2006.
!*   All Rights Reserved.
!*   US Government Users Restricted Rights - 
!*   Use, duplication or disclosure restricted by 
!*   GSA ADP Schedule Contract with IBM Corp.
!*
!***********************************************************************
!
!-----------------------------------------------------------------------
! * NAME: xl__ieee
! *
! * COMPONENT_NAME: SIGFPE or SIGTRAP signal handler for floating-point
! *                 exceptions
! *
! * FUNCTION: exception handler to simulate a "trap disabled" IEEE 754
! *           environment
! *
! * ORIGINS: 27
! *
! * NOTES:
! *
! * This code is a SIGFPE or SIGTRAP signal handler for floating-point
! * exceptions (SIGTRAP signals are generated for floating-point
! * exceptions when the 'FLTTRAP' compiler option is used).  The handler
! * includes the ability to identify the point where an exception occurred
! * and to continue execution following an exception, supplying a value as
! * the result of the failing instruction.
! *
! * This version of the handler attempts to supply the default IEEE 754
! * result for any operation that generates a floating-point exception.
! * Normally, when floating-point exceptions are enabled it is not
! * possible to continue execution following an exception, because the
! * value (if any) produced by the failing instruction is not appropriate
! * for further processing.  This exception handler allows the points
! * where exceptions occur during execution to be identified, while also
! * allowing execution to proceed as if exceptions were not enabled.
! *
! * The name of the handler is 'xl__ieee'.
! *
! * This source code should be compiled using 'xlf -c -O xl__ieee.F'.
! *
! * This code may be further customized to print additional debugging
! * information or define other exception results.
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!            NOTICE TO USERS OF THE SOURCE CODE EXAMPLES
!
! THE SOURCE CODE EXAMPLES PROVIDED BY IBM ARE ONLY INTENDED TO ASSIST
! IN THE DEVELOPMENT OF A WORKING SOFTWARE PROGRAM.  THE SOURCE CODE
! EXAMPLES DO NOT FUNCTION AS WRITTEN:  ADDITIONAL CODE IS REQUIRED.
! IN ADDITION, THE SOURCE CODE EXAMPLES MAY NOT COMPILE AND/OR BIND
! SUCCESSFULLY AS WRITTEN.
!
! INTERNATIONAL BUSINESS MACHINES CORPORATION PROVIDES THE SOURCE CODE
! EXAMPLES, BOTH INDIVIDUALLY AND AS ONE OR MORE GROUPS, "AS IS" WITHOUT
! WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT
! LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE
! OF THE SOURCE CODE EXAMPLES, BOTH INDIVIDUALLY AND AS ONE OR MORE
! GROUPS, IS WITH YOU.  SHOULD ANY PART OF THE SOURCE CODE EXAMPLES
! PROVE DEFECTIVE, YOU (AND NOT IBM OR AN AUTHORIZED RISC System/6000*
! WORKSTATION DEALER) ASSUME THE ENTIRE COST OF ALL NECESSARY SERVICING,
! REPAIR OR CORRECTION.
!
! IBM does not warrant that the contents of the source code examples,
! whether individually or as one or more groups, will meet your
! requirements or that the source code examples are error-free.
!
! IBM may make improvements and/or changes in the source code examples
! at any time.
!
! Changes may be made periodically to the information in the source code
! examples; these changes may be reported, for the sample exception
! handler included herein, in new editions of the examples.
!
! References in the source code examples to IBM products, programs, or
! services do not imply that IBM intends to make these available in all
! countries in which IBM operates.  Any reference to an IBM licensed
! program in the source code examples is not intended to state or imply
! that only IBM's licensed program may be used.  Any functionally
! equivalent program may be used.
!
! * RISC System/6000 is a trademark of International Business Machines
!   Corporation.
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! * The FLTTRAP compiler option generates SIGTRAP signals when
! * floating-point exceptions occur.  It does so by setting the
! * record bit on all floating-point instructions, and then
! * trapping if condition register bit 5 is set (that is, if the
! * floating-point enabled exception (FEX) bit is set in the
! * floating-point status and control register (FPSCR)).
! *
! * To use this handler for SIGTRAP signals, the user code must be
! * compiled with the FLTTRAP compiler option, must enable the desired
! * floating-point exceptions (possibly with the FLTTRAP 'enable'
! * suboption), and must install the 'xl__ieee' routine as the
! * SIGTRAP signal handler (possibly with the 'sigtrap' compiler
! * option).  For example, to detect invalid operation exceptions
! * in a program called 'invop.f', use the commands:
! *     xlf -c -O xl__ieee.F
! *     xlf -qflttrap=en:inv -qsigtrap=xl__ieee invop.f xl__ieee.o
! *
! * To use this handler for SIGFPE signals, the user code must
! * enable the desired floating-point exceptions, must run in
! * synchronous mode, and must install the 'xl__ieee' routine as the
! * SIGFPE signal handler.  The user code should NOT be compiled with
! * the FLTTRAP compiler option.  For example, to detect invalid
! * operation exceptions, the following code could be added to the
! * startup code of a Fortran program:
! *     external xl__ieee
! *     call fp_enable(%val('00000080'x))   ! TRP_INVALID
! *     call fp_trap(%val(1))               ! FP_TRAP_SYNC
! *     call signal(8, xl__ieee)            ! SIGFPE
! * The xl__ieee code would then be compiled and linked with the user
! * code as shown above.
! * The 'fp_enable' and 'fp_trap' routines are documented entry points
! * in the standard 'C' library in AIX.  Hexadecimal parameter values
! * for the routines are from the include file '/usr/include/fptrap.h'.
! *
! * For complete information on the floating-point processor in
! * the POWER, POWER2 and PowerPC architectures, and as an aid in
! * understanding the operation of this signal handler, refer to:
! * "The PowerPC Architecture: A Specification for a New Family of
! * RISC Processors", Morgan Kaufmann Publishing, Inc., Second
! * Edition, 1994.
!-----------------------------------------------------------------------

@PROCESS DEBUG(CALLBY)

#if defined(__linux__) || defined(__APPLE__)
      subroutine xl__ieee (%val(signal), info, sigcontextdata)
#include "fsignal.h"

        type(siginfo_t)  :: info
#if defined(__APPLE__) && defined(__64BIT__)
        type(ucontext64_t) :: sigcontextdata
#else
        type(ucontext_t) :: sigcontextdata
#endif

#else
      subroutine xl__ieee (%val(signal), %val(code), sigcontextdata)
#include "fsignal.h"

        __INT32 code
        type(sigcontext) :: sigcontextdata
#endif
        include 'fexcp.h'
        __INT32 signal

        ! The specific trap instruction used by the FLTTRAP option is
        ! TRAP R15=R15.  This is the machine code for that instruction.
        integer, parameter :: FLTTRAP_INST    = x'7C8F7808'

        ! Preceding the trap instruction, the FLTTRAP option generates
        ! a conditional branch based on condition register bit 5.
        ! This is the machine code for that instruction.
        integer, parameter :: FLTTRAP_BRANCH  = x'40850008'

        ! The following masks for the exception enable bits in the
        ! FPSCR are from /usr/include/fptrap.h.
        integer, parameter :: TRP_INVALID     = x'00000080'
        integer, parameter :: TRP_OVERFLOW    = x'00000040'
        integer, parameter :: TRP_UNDERFLOW   = x'00000020'
        integer, parameter :: TRP_DIV_BY_ZERO = x'00000010'
        integer, parameter :: TRP_INEXACT     = x'00000008'

        ! The following masks for the exception indication bits in
        ! the FPSCR are from /usr/include/fpxcp.h.
        integer, parameter :: FP_INVALID      = x'20000000'
        integer, parameter :: FP_OVERFLOW     = x'10000000'
        integer, parameter :: FP_UNDERFLOW    = x'08000000'
        integer, parameter :: FP_DIV_BY_ZERO  = x'04000000'
        integer, parameter :: FP_INEXACT      = x'02000000'
        integer, parameter :: FP_INV_SNAN     = x'01000000'

        ! FP_ALL_XCP, from /usr/include/fpxcp.h, has set bits for all
        ! of the floating-point status and control register bits that
        ! indicate that an exception has occurred and that must be
        ! cleared before resuming execution.
        integer, parameter :: FP_ALL_XCP = x'BFF80700'

        ! FP_FPRF_MASK has set bits corresponding to the result flags
        ! field of the FPSCR.
        ! Other values are result flags for particular result types
        ! (See Figure 25 of the PowerPC Architecture manual).
        integer, parameter :: FP_FPRF_MASK    = x'0001F000'
        integer, parameter :: FP_FPRF_QNAN    = x'00011000'
        integer, parameter :: FP_FPRF_NINF    = x'00009000'
        integer, parameter :: FP_FPRF_NNORM   = x'00008000'
        integer, parameter :: FP_FPRF_NDEN    = x'00018000'
        integer, parameter :: FP_FPRF_NZERO   = x'00012000'
        integer, parameter :: FP_FPRF_PZERO   = x'00002000'
        integer, parameter :: FP_FPRF_PDEN    = x'00014000'
        integer, parameter :: FP_FPRF_PNORM   = x'00004000'
        integer, parameter :: FP_FPRF_PINF    = x'00005000'

        ! FP_FR and FP_FI are masks to manipulate the FR and FI bits
        ! in the FPSCR.
        integer, parameter :: FP_FR           = x'00040000'
        integer, parameter :: FP_FI           = x'00020000'

        ! The fpstat variable is used as a local copy of the FPSCR
        ! value at the point where the signal occurred.

        ! The 'instr' variable will be used to examine the code at the
        ! point where the signal was generated.
        integer(4) fpstat
        integer(4) instr

        pointer (code_ptr,instr)

        ! These variables are used when the instruction that caused the
        ! exception is decoded.

        ! The 'sign', 'signif' and 'expon' variables are used when it is
        ! necessary to extract and manipulate the fields of a double
        ! precision register value.  The 'G' and 'X' variables are used
        ! for denormalization of underflow results.
        integer(4) opcd, xo, frt, fra, frb, frc  ! instruction fields
        integer(4) num_sources, source(3), src
        integer(8) sign, signif
        integer(4) expon, G, X

        ! The 'ifpr' variable will be used to examine the saved floating-
        ! point registers in the sigcontext structure as binary values.
        integer(8) ifpr(0:31)
        pointer (fpr_ptr,ifpr)

        integer(4) fpscr
        pointer(fpscr_ptr, fpscr)

        integer(4) nip ! next instruction pointer, also called iar
        pointer(nip_ptr, nip)

#ifdef __linux__
        ! The 'regs' variable will be used to access the target of the
        ! regs pointer in the sigcontext structure on linux.
        type(pt_regs) :: regs
        pointer(regs_ptr,regs)
#endif

        ! Executable code for the signal handler begins here.

        ! Call the routine xl__sigdump, from libxlf90.a, to print
        ! information about the signal that was received and
        ! a traceback starting from the point where the signal
        ! was generated.
#ifdef _AIX
        call xl__sigdump (%val(signal), %val(code), sigcontextdata)
#else
        call xl__sigdump (%val(signal), info, sigcontextdata)
#endif
        ! Prepare for a return to the instruction following the point
        ! where the signal was generated.
        ! First, make local copies of the floating-point status and
        ! control register and the iar register from the point where
        ! the exception occurred.
#ifdef _AIX
        fpscr_ptr = loc(sigcontextdata%sc_jmpbuf%jmp_context%fpscr)
        fpstat = fpscr
        nip_ptr = loc(sigcontextdata%sc_jmpbuf%jmp_context%iar)
        code_ptr = sigcontextdata%sc_jmpbuf%jmp_context%iar
#elif defined(__linux__)
        regs_ptr = sigcontextdata%uc_mcontext%regs
        fpscr_ptr = loc(regs%gpr(PT_FPSCR))
        fpstat = fpscr
        nip_ptr = loc(regs%gpr(PT_NIP))
        code_ptr = regs%gpr(PT_NIP)
#elif defined(__APPLE__)
#if defined(__64BIT__)
        fpscr_ptr = loc(sigcontextdata%uc_mcontext64%fs%fpscr)
        fpstat = fpscr
        nip_ptr = loc(sigcontextdata%uc_mcontext64%ss%srr0)
        code_ptr = sigcontextdata%uc_mcontext64%ss%srr0
#else
        fpscr_ptr = loc(sigcontextdata%uc_mcontext%fs%fpscr)
        fpstat = fpscr
        nip_ptr = loc(sigcontextdata%uc_mcontext%ss%srr0)
        code_ptr = sigcontextdata%uc_mcontext%ss%srr0
#endif
#endif

        ! Next, clear all exception bits in the saved floating-point
        ! status and control register; this will prevent future
        ! signals due to floating-point exceptions that occurred
        ! before this signal.
        fpscr = and(fpscr, not(FP_ALL_XCP))
        ! When a signal is received, the saved iar points at the failing
        ! instruction; increment the iar so that the return from the
        ! signal handler does not immediately fail again.
        nip = nip + 4


        if (signal .eq. SIGTRAP) then
          ! Check that the trap is of the type used for the FLTTRAP
          ! option.  If it is not, terminate execution (other trap
          ! causes, such as an integer division by zero or a subscript
          ! out of range, are best treated as fatal errors).
          if (instr .ne. FLTTRAP_INST)                                  &
     &      STOP '(xl__ieee: not a FLTTRAP exception)'

          ! Examine the code sequence leading to the trap to ensure it
          ! was generated by the FLTTRAP option.
          ! First, ensure the instruction preceding the trap is the expected
          ! conditional branch generated by the FLTTRAP option.
          code_ptr = code_ptr - 4
          if (instr .ne. FLTTRAP_BRANCH)                                &
     &      STOP '(xl__ieee: invalid FLTTRAP code sequence)'
          ! The instruction before the branch should be a floating-point
          ! operation with record mode.
          code_ptr = code_ptr - 4
          ! Make sure the record bit is set.
          if (and(instr,1) .ne. 1)                                      &
     &      STOP '(xl__ieee: invalid FLTTRAP code sequence)'
        elseif (signal .ne. SIGFPE) then
          STOP '(xl__ieee: not a SIGTRAP or SIGFPE signal)'
        end if

        ! Floating-point operations have an opcode field value of 63
        ! (double precision) or 59 (single precision).
        opcd = ishft(instr,-26)
        if ((opcd .ne. 63) .and. (opcd .ne. 59))                        &
     &    STOP '(xl__ieee: exception point cannot be determined)'

        ! Decode the XO field of the instruction, assuming it is
        ! A-form.
        xo = and(ishft(instr,-1),x'1F')

        ! Check to see if the instruction found was a move register (fmr).
        if (xo .eq. 8) then  ! move register
          ! A move register is used to detect a floating-point
          ! exception that occurred in a previous instruction.
          ! This is done after calls to external routines (to see if
          ! they returned with any exception bits set), after
          ! floating-point compares (which do not have a record mode
          ! form), and when the 'imprecise' suboption of FLTTRAP is used.
          ! For compares, the compare operation will be just before the
          ! move register.
          code_ptr = code_ptr - 4
          opcd = ishft(instr,-26)
          xo = and(ishft(instr,-1),x'1F')
          if ((opcd .ne. 63) .or. (xo .ne. 0)) then  ! not a compare
            ! For calls, any external routine that generated an exception
            ! but did not generate a SIGTRAP must be a library routine
            ! or be in user code that was not compiled with FLTTRAP.  For
            ! those cases, terminate execution because the original failing
            ! operation cannot be found.
            ! Similarly, if the 'imprecise' FLTTRAP mode was used, the
            ! failing operation cannot be found.
            STOP '(xl__ieee: exception point cannot be determined)'
          ! else: found the floating point compare that caused the exception
          end if
        end if

        ! Decode the rest of the instruction.
        frt = and(ishft(instr,-21),x'1F')
        fra = and(ishft(instr,-16),x'1F')
        frb = and(ishft(instr,-11),x'1F')
        frc = and(ishft(instr,-6),x'1F')

        ! Use the opcode and register information to analyze the
        ! failing instruction.
        select case(xo)
          case(21)  ! add operation
            num_sources = 2
            source(1) = fra
            source(2) = frb
          case(20)  ! subtract operation
            num_sources = 2
            source(1) = fra
            source(2) = frb
          case(25)  ! multiply operation
            num_sources = 2
            source(1) = fra
            source(2) = frc
          case(18)  ! divide operation
            num_sources = 2
            source(1) = fra
            source(2) = frb
          case(29)  ! multiply-add operation
            num_sources = 3
            source(1) = fra
            source(2) = frb
            source(3) = frc
          case(28)  ! multiply-subtract operation
            num_sources = 3
            source(1) = fra
            source(2) = frb
            source(3) = frc
          case(31)  ! negative multiply-add operation
            num_sources = 3
            source(1) = fra
            source(2) = frb
            source(3) = frc
          case(30)  ! negative multiply-subtract operation
            num_sources = 3
            source(1) = fra
            source(2) = frb
            source(3) = frc
          case(12)  ! round to single-precision operation
            num_sources = 1
            source(1) = frb
          case(14)  ! convert to/from integer
            num_sources = 1
            source(1) = frb
          case(15)  ! convert to integer with round toward zero
            num_sources = 1
            source(1) = frb
          case(0)  ! compare operation
            ! Compares can only generate invalid operation exceptions, and
            ! do so when one of the operands of the compare is a NaN.
            ! No fixup is needed (the condition register result is set to
            ! indicate unordered), so just continue execution.
            return
          case(22)  ! square root
            num_sources = 1
            source(1) = frb
          case(24)  ! reciprocal estimate
            num_sources = 1
            source(1) = frb
          case(26)  ! reciprocal square root estimate
            num_sources = 1
            source(1) = frb
          case default
            ! This could be an exception caused by explicitly setting
            ! bits in the floating-point status and control register.
            STOP '(xl__ieee: exception point cannot be determined)'
        end select

        ! Clear the FPRF field of the FPSCR in preparation for setting
        ! it to indicate the class of the result returned below.
        fpscr = and(fpscr, not(FP_FPRF_MASK))

#ifdef _AIX
        ! Examine the floating point status and control register for
        ! enabled exceptions, and set the result register to the
        ! IEEE default result.
        fpr_ptr = loc(sigcontextdata%sc_jmpbuf%jmp_context%fpr)
#elif defined(__linux__)
        fpr_ptr = loc(regs%gpr(PT_FPR0))
#elif defined(__APPLE__)
#if defined(__64BIT__)
        fpr_ptr = loc(sigcontextdata%uc_mcontext64%fs%fpregs)
#else
        fpr_ptr = loc(sigcontextdata%uc_mcontext%fs%fpregs)
#endif
#endif

        if (and(fpstat,or(TRP_INVALID,FP_INVALID)) .eq.                 &
     &      or(TRP_INVALID,FP_INVALID)) then
          ! The default result for a disabled invalid operation depends
          ! on the instruction.
          if ((xo .eq. 14) .or. (xo .eq. 15)) then  ! convert to integer
            ! If the source operand is negative or a NaN, the result
            ! is the most negative integer, else the result is the
            ! the most positive integer.
            if ((and(ifpr(frb),z'8000000000000000') .ne. 0) .or.        &
     &          ((and(ifpr(frb),z'7FF0000000000000') .eq.               &
     &              z'7FF0000000000000') .and.                          &
     &           (and(ifpr(frb),z'000FFFFFFFFFFFFF') .ne. 0))) then
              if (frc .eq. 0) then  ! 32-bit result
                ifpr(frt) = z'0000000080000000'
              else
                ifpr(frt) = z'8000000000000000'
              end if
            else
              if (frc .eq. 0) then  ! 32-bit result
                ifpr(frt) = z'000000007FFFFFFF'
              else
                ifpr(frt) = z'7FFFFFFFFFFFFFFF'
              end if
            end if
            ! The FPRF field of the FPSCR is undefined for these conversions.
          else  ! not a convert to integer operation
            ! For all other cases, the result is a quiet NaN.  If one
            ! of the source operands is a signaling NaN, the result
            ! should be a quiet version of that NaN; otherwise, the
            ! result is a generic quiet NaN.
            if (and(fpstat,FP_INV_SNAN) .ne. 0) then
              do src=1,num_sources
                if ((and(ifpr(source(src)),z'7FF8000000000000') .eq.    &
     &                  z'7FF0000000000000') .and.                      &
     &              (and(ifpr(source(src)),z'000FFFFFFFFFFFFF') .ne.    &
     &                  0)) then
                  ifpr(frt) = or(ifpr(source(src)),z'0008000000000000')
                  exit
                end if
              end do
              if (xo .eq. 12) then  ! round to single-precision operation
                ! Truncate the fraction to produce a valid single-precision
                ! result.
                ifpr(frt) = and(ifpr(frt),z'FFFFFFFFE0000000')
              end if
            else
              ifpr(frt) = z'7ff8000000000000'  ! generic quiet nan
            end if
            ! Set the FPRF field of the FPSCR to indicate a quiet NaN result.
            fpscr = or(fpscr, FP_FPRF_QNAN)
          end if
        ! End of invalid operation processing code.

        elseif (and(fpstat,or(TRP_DIV_BY_ZERO,FP_DIV_BY_ZERO)) .eq.     &
     &      or(TRP_DIV_BY_ZERO,FP_DIV_BY_ZERO)) then
          ! The default result is infinity, where the sign is determined
          ! by the XOR of the signs of the operands.
          ifpr(frt) = or(z'7FF0000000000000',                           &
     &           and(xor(ifpr(fra),ifpr(frb)),z'8000000000000000'))
          ! Set the FPRF field of the FPSCR to indicate an infinity result.
          if (and(ifpr(frt),z'8000000000000000') .eq. 0) then
            fpscr = or(fpscr, FP_FPRF_PINF)
          else
            fpscr = or(fpscr, FP_FPRF_NINF)
          end if
        ! End of divide by zero processing code.

        elseif (and(fpstat,or(TRP_OVERFLOW,FP_OVERFLOW)) .eq.           &
     &      or(TRP_OVERFLOW,FP_OVERFLOW)) then
          ! The result register in an overflow contains a rounded
          ! normalized number, but 1536 (double precision) or 192
          ! (single precision) has been subtracted from the exponent.
          ! The default result depends on the rounding mode and sign
          ! of the result.
          sign = and(ifpr(frt),z'8000000000000000')
          select case(and(fpstat,3))
            case(0)  ! round to nearest
              ifpr(frt) = or(z'7FF0000000000000',sign)  ! +/- infinity
              if (sign .eq. 0) then
                fpscr = or(fpscr, FP_FPRF_PINF)
              else
                fpscr = or(fpscr, FP_FPRF_NINF)
              end if
            case(1)  ! round toward zero
              if ((opcd .eq. 59) .or. (xo .eq. 12)) then  ! single precision
                ifpr(frt) = z'7F7FFFFF'
              else
                ifpr(frt) = z'7FEFFFFFFFFFFFFF'
              end if
              ifpr(frt) = or(ifpr(frt),sign)
              if (sign .eq. 0) then
                fpscr = or(fpscr, FP_FPRF_PNORM)
              else
                fpscr = or(fpscr, FP_FPRF_NNORM)
              end if
            case(2)  ! round toward +infinity
              if (sign .eq. 0) then
                ifpr(frt) = z'7FF0000000000000'  ! +infinity
                fpscr = or(fpscr, FP_FPRF_PINF)
              else
                if ((opcd .eq. 59) .or. (xo .eq. 12)) then  ! single precision
                  ifpr(frt) = z'FF7FFFFF'
                else
                  ifpr(frt) = z'FFEFFFFFFFFFFFFF'
                end if
                fpscr = or(fpscr, FP_FPRF_NNORM)
              end if
            case(3)  ! round toward -infinity
              if (sign .eq. 0) then
                if ((opcd .eq. 59) .or. (xo .eq. 12)) then  ! single precision
                  ifpr(frt) = z'7F7FFFFF'
                else
                  ifpr(frt) = z'7FEFFFFFFFFFFFFF'
                end if
                fpscr = or(fpscr, FP_FPRF_PNORM)
              else
                ifpr(frt) = z'FFF0000000000000'  ! -infinity
                fpscr = or(fpscr, FP_FPRF_NINF)
              end if
          end select
          ! The FPSCR FR bit is undefined, and the FI bit should be set.
          fpscr = or(fpscr, FP_FI)
        ! End of overflow processing code.

        elseif (and(fpstat,or(TRP_UNDERFLOW,FP_UNDERFLOW)) .eq.         &
     &      or(TRP_UNDERFLOW,FP_UNDERFLOW)) then
          ! The result register in an underflow contains a rounded
          ! normalized number, but 1536 (double precision) or 192
          ! (single precision) has been added to the exponent.
          ! The default result depends on the rounding mode and sign
          ! of the result.
          sign = and(ifpr(frt),z'8000000000000000')
          expon = and(ishft(ifpr(frt),-52_8),z'7FF')
          signif = or(and(ifpr(frt),z'000FFFFFFFFFFFFF'),               &
     &              z'0010000000000000')
          if ((opcd .eq. 59) .or. (xo .eq. 12)) then  ! single precision
            expon = expon - 192
            ! Shift the double precision significand 29 bits right to get
            ! the short precision significand.
            signif = ishft(signif,-29_8)
          else
            expon = expon - 1536
          end if
          ! The 'expon' value is now a negative number, and represents the
          ! right shift amount that must be applied to the significand
          ! (that is, by shifting the significand and incrementing the
          ! exponent until the exponent has value one, we can derive the
          ! fraction for the default result).
          ! The FR and FI bits in the FPSCR allow us to 'unround' the
          ! result produced.  If set, the FI bit indicates that some
          ! type of rounding was applied, depending on the rounding
          ! mode and the G, R and X bits.  If the FR bit is set, the
          ! significand was incremented as part of the rounding.
          ! For our purposes, we will need to maintain a version of the
          ! guard (G) bit, and a sticky (X) bit.  The round (R) bit does
          ! not need to be explicitly maintained, so our X bit will
          ! serve as a logical OR of the R and X bits specified by an
          ! IEEE execution model.
          ! Initially, assume no rounding.
          G = 0
          X = 0
          if (and(fpstat,FP_FI) .ne. 0) then
            ! Some rounding was applied, depending on the rounding mode.
            ! For our purposes, it is sufficient to assume that the X
            ! bit should be set.
            X = 1
            ! If the significand was incremented as part of rounding, then
            ! decrement it now and set the G bit, which is again sufficient
            ! for our purposes.
            if (and(fpstat,FP_FR) .ne. 0) then
              signif = signif - 1
              G = 1
              ! Clear the FR bit in the saved FPSCR; it will be set below
              ! based on the rounding of the default result.
              fpscr = or(fpscr, not(FP_FR))
            end if
          end if
          ! Now shift the significand right to denormalize the result.
          ! For efficiency, stop shifting if both the significand and
          ! the guard bit are zero.
          do while ((expon .le. 0) .and.                                &
     &        ((signif .ne. 0) .or. (G .ne. 0)))
            X = or(X,G)
            G = and(signif,1_8)
            signif = ishft(signif,-1_8)
            expon = expon + 1
          end do
          if ((G .ne. 0) .or. (X .ne. 0)) then
            ! Result is inexact.  Make sure FI is set in the FPSCR, and
            ! use to rounding mode to complete the rounding of the
            ! denormalized result.
            fpscr = or(fpscr, FP_FI)
            select case(and(fpstat,3))
              case(0)  ! round to nearest
                if ((G .ne. 0) .and.                                    &
     &              ((X .ne. 0) .or. (and(signif,1_8) .ne. 0))) then
                  signif = signif + 1
                  fpscr = or(fpscr, FP_FR)
                end if
              case(1)  ! round toward zero
              case(2)  ! round toward +infinity
                if (sign .eq. 0) then
                  signif = signif + 1
                  fpscr = or(fpscr, FP_FR)
                end if
              case(3)  ! round toward - infinity
                if (sign .ne. 0) then
                  signif = signif + 1
                  fpscr = or(fpscr, FP_FR)
                end if
            end select
          end if
          if ((opcd .eq. 59) .or. (xo .eq. 12)) then  ! single precision
            ! Shift the short precision significand 29 bits left to get
            ! a double precision significand for the result register.
            signif = ishft(signif,29_8)
          end if
          ! The final result is the original sign, a zero exponent, and
          ! the derived significand.
          ifpr(frt) = or(sign,signif)
          if (sign .eq. 0) then
            if (signif .eq. 0) then
              fpscr = or(fpscr, FP_FPRF_PZERO) ! result is positive zero
            else
              fpscr = or(fpscr, FP_FPRF_PDEN)  ! result is a positive denormal
            end if
          else
            if (signif .eq. 0) then
              fpscr = or(fpscr, FP_FPRF_NZERO) ! result is negative zero
            else
              fpscr = or(fpscr, FP_FPRF_NDEN)  ! result is a negative denormal
            end if
          end if
        ! End of underflow processing code.

        elseif (and(fpstat,or(TRP_INEXACT,FP_INEXACT)) .eq.             &
     &      or(TRP_INEXACT,FP_INEXACT)) then
          ! No action necessary, just ignore this.
        end if

      end
