/**********************************************************************/
/*                                                                    */
/*   Licensed Materials - Property of IBM.                            */
/*   IBM XL Fortran Advanced Edition V10.1 for Linux (5724-M17)       */
/*   Copyright IBM Corp. 2003, 2006.                                  */
/*   All Rights Reserved.                                             */
/*   US Government Users Restricted Rights -                          */
/*   Use, duplication or disclosure restricted by                     */
/*   GSA ADP Schedule Contract with IBM Corp.                         */
/*                                                                    */
/**********************************************************************/

/* ---------------------------------------------------------------------
! * NAME: xl__ieee
! *
! * COMPONENT_NAME: SIGFPE or SIGTRAP signal handler for floating-point
! *                 exceptions
! *
! * FUNCTION: exception handler to simulate a "trap disabled" IEEE 754
! *           environment
! *
! * ORIGINS: 27
! *
! * NOTES:
! *
! * This code is a SIGFPE or SIGTRAP signal handler for floating-point
! * exceptions (When the 'FLTTRAP' compiler option is used on AIX, SIGTRAP
! * signals are generated for floating-point exceptions.  When it is used
! * on Linux PPC, SIGFPE signals are generated.)  The handler includes the
! * ability to identify the point where an exception occurred and to
! * continue execution following an exception, supplying a value as
! * the result of the failing instruction.
! *
! * This version of the handler attempts to supply the default IEEE 754
! * result for any operation that generates a floating-point exception.
! * Normally, when floating-point exceptions are enabled it is not
! * possible to continue execution following an exception, because the
! * value (if any) produced by the failing instruction is not appropriate
! * for further processing.  This exception handler allows the points
! * where exceptions occur during execution to be identified, while also
! * allowing execution to proceed as if exceptions were not enabled.
! *
! * The name of the handler is 'xl__ieee'.
! *
! * This source code should be compiled using 'cc -c -O xl__ieee.c'.
! * Do not compile with 'xlc', because this code violates the ANSI C
! * aliasing restrictions.
! *
! * This code may be further customized to print additional debugging
! * information or define other exception results.
!-------------------------------------------------------------------- */

/* ---------------------------------------------------------------------
!            NOTICE TO USERS OF THE SOURCE CODE EXAMPLES
!
! THE SOURCE CODE EXAMPLES PROVIDED BY IBM ARE ONLY INTENDED TO ASSIST
! IN THE DEVELOPMENT OF A WORKING SOFTWARE PROGRAM.  THE SOURCE CODE
! EXAMPLES DO NOT FUNCTION AS WRITTEN:  ADDITIONAL CODE IS REQUIRED.
! IN ADDITION, THE SOURCE CODE EXAMPLES MAY NOT COMPILE AND/OR BIND
! SUCCESSFULLY AS WRITTEN.
!
! INTERNATIONAL BUSINESS MACHINES CORPORATION PROVIDES THE SOURCE CODE
! EXAMPLES, BOTH INDIVIDUALLY AND AS ONE OR MORE GROUPS, "AS IS" WITHOUT
! WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT
! LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE
! OF THE SOURCE CODE EXAMPLES, BOTH INDIVIDUALLY AND AS ONE OR MORE
! GROUPS, IS WITH YOU.  SHOULD ANY PART OF THE SOURCE CODE EXAMPLES
! PROVE DEFECTIVE, YOU (AND NOT IBM OR AN AUTHORIZED RISC System/6000*
! WORKSTATION DEALER) ASSUME THE ENTIRE COST OF ALL NECESSARY SERVICING,
! REPAIR OR CORRECTION.
!
! IBM does not warrant that the contents of the source code examples,
! whether individually or as one or more groups, will meet your
! requirements or that the source code examples are error-free.
!
! IBM may make improvements and/or changes in the source code examples
! at any time.
!
! Changes may be made periodically to the information in the source code
! examples; these changes may be reported, for the sample exception
! handler included herein, in new editions of the examples.
!
! References in the source code examples to IBM products, programs, or
! services do not imply that IBM intends to make these available in all
! countries in which IBM operates.  Any reference to an IBM licensed
! program in the source code examples is not intended to state or imply
! that only IBM's licensed program may be used.  Any functionally
! equivalent program may be used.
!
! * RISC System/6000 is a trademark of International Business Machines
!   Corporation.
!-------------------------------------------------------------------- */

/* ---------------------------------------------------------------------
! * The FLTTRAP compiler option generates SIGTRAP signals on AIX
! * and SIGFPE signal on linux PPC when floating-point exceptions
! * occur.  It does so by setting the record bit on all floating-point
! * instructions, and then trapping if condition register bit 5 is set
! * (that is, if the floating-point enabled exception (FEX) bit is set
! * in the floating-point status and control register (FPSCR)).
! *
! * On AIX:
! *
! * To use this handler for SIGTRAP signals, the user code must be
! * compiled with the FLTTRAP compiler option, must enable the desired
! * floating-point exceptions (possibly with the FLTTRAP 'enable'
! * suboption), and must install the 'xl__ieee' routine as the
! * SIGTRAP signal handler (possibly with the 'sigtrap' compiler
! * option).  For example, to detect invalid operation exceptions
! * in a program called 'invop.f', use the commands:
! *     cc -c -O xl__ieee.c
! *     xlf -qflttrap=en:inv -qsigtrap=xl__ieee invop.f xl__ieee.o
! *
! * To use this handler for SIGFPE signals, the user code must
! * enable the desired floating-point exceptions, must run in
! * synchronous mode, and must install the 'xl__ieee' routine as the
! * SIGFPE signal handler.  The user code should NOT be compiled with
! * the FLTTRAP compiler option.  For example, to detect invalid
! * operation exceptions, the following code could be added to the
! * startup code of a Fortran program:
! *     external xl__ieee
! *     call fp_enable(%val('00000080'x))   ! TRP_INVALID
! *     call fp_trap(%val(1))               ! FP_TRAP_SYNC
! *     call signal(8, xl__ieee)            ! SIGFPE
! * The xl__ieee code would then be compiled and linked with the user
! * code as shown above.
! * The 'fp_enable' and 'fp_trap' routines are documented entry points
! * in the standard 'C' library in AIX.  Hexadecimal parameter values
! * for the routines are from the include file '/usr/include/fptrap.h'.
! *
! * On Linux and MacOS X PPC:
! *
! * To use this handler for SIGFPE signals, the user code must be
! * compiled with the FLTTRAP compiler option, must enable the desired
! * floating-point exceptions (possibly with the FLTTRAP 'enable'
! * suboption), and must install the 'xl__ieee' routine as the
! * SIGFPE signal handler (possibly with the 'sigtrap' compiler
! * option).  For example, to detect invalid operation exceptions
! * in a program called 'invop.f', use the commands:
! *     cc -c -O xl__ieee.c
! *     xlf -qflttrap=en:inv -qsigtrap=xl__ieee invop.f xl__ieee.o
! *
! *
! * For complete information on the floating-point processor in
! * the POWER, POWER2 and PowerPC architectures, and as an aid in
! * understanding the operation of this signal handler, refer to:
! * "The PowerPC Architecture: A Specification for a New Family of
! * RISC Processors", Morgan Kaufmann Publishing, Inc., Second
! * Edition, 1994.
!-------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/ucontext.h>

#if !defined(__linux__) && !defined(__APPLE__)

#include <fptrap.h>
#include <fpxcp.h>

#else

typedef unsigned int fptrap_t;

/* ******** ENABLE EXCEPTION DETAIL BITS ***********/

#define TRP_INVALID     ((fptrap_t) 0x00000080)
#define TRP_OVERFLOW    ((fptrap_t) 0x00000040)
#define TRP_UNDERFLOW   ((fptrap_t) 0x00000020)
#define TRP_DIV_BY_ZERO ((fptrap_t) 0x00000010)
#define TRP_INEXACT     ((fptrap_t) 0x00000008)

typedef unsigned int fpflag_t;

/*  ******* REQUIRED SUMMARY BITS *********** */

#define FP_INVALID      ((fpflag_t) 0x20000000)
#define FP_OVERFLOW     ((fpflag_t) 0x10000000)
#define FP_UNDERFLOW    ((fpflag_t) 0x08000000)
#define FP_DIV_BY_ZERO  ((fpflag_t) 0x04000000)
#define FP_INEXACT      ((fpflag_t) 0x02000000)

/*  ******* MACHINE SPECIFIC SUMMARY BITS *********** */

#define FP_ALL_XCP      ((fpflag_t) 0xBFF80700) /* Bits 1 and 2 cannot
                                                 * be (re)set */

/* ******** MACHINE SPECIFIC INVALID EXCEPTION DETAIL BITS ***********/

#define FP_INV_SNAN     ((fpflag_t) 0x01000000) /* Signalling NaN */

#endif /* __linux__ && __APPLE__ */

/* The specific trap instruction used by the FLTTRAP option is
 * TRAP R15=R15.  This is the machine code for that instruction. */
#define FLTTRAP_INST    (0x7C8F7808)

/* Preceding the trap instruction, the FLTTRAP option generates
 * a conditional branch based on condition register bit 5.
 * This is the machine code for that instruction. */
#define FLTTRAP_BRANCH  (0x40850008)

/* For -qflttrap=nanq a different branch instruction precedes the
 * trap instruction.  This is the machine code for that instruction. */
#define FLTTRAP_NANQ_BRANCH (0x41E60008)

/* FP_FPRF_MASK has set bits corresponding to the result flags
 * field of the FPSCR.
 * Other values are result flags for particular result types
 * (See Figure 25 of the PowerPC Architecture manual). */
#define FP_FPRF_MASK    (0x0001F000)
#define FP_FPRF_QNAN    (0x00011000)
#define FP_FPRF_NINF    (0x00009000)
#define FP_FPRF_NNORM   (0x00008000)
#define FP_FPRF_NDEN    (0x00018000)
#define FP_FPRF_NZERO   (0x00012000)
#define FP_FPRF_PZERO   (0x00002000)
#define FP_FPRF_PDEN    (0x00014000)
#define FP_FPRF_PNORM   (0x00004000)
#define FP_FPRF_PINF    (0x00005000)

/* FP_FR and FP_FI are masks to manipulate the FR and FI bits
 * in the FPSCR. */
#define FP_FR           (0x00040000)
#define FP_FI           (0x00020000)

/* The following constants represent some special floating
 * point values.  Single precision values are converted to
 * double precision since intermediate results are kept in
 * 8-byte registers.  Therefore:
 * The largest finite single is 0x7F7FFFFF (single), which
 * is 0x47EFFFFFE0000000 (double).
 * The smallest finite single is 0xFF7FFFFF (single), which
 * is 0xC7EFFFFFE0000000 (double). */
#define LARGEST_FINITE_SINGLE   (0x47EFFFFFE0000000LL)
#define SMALLEST_FINITE_SINGLE  (0xC7EFFFFFE0000000LL)
#define LARGEST_FINITE_DOUBLE   (0x7FEFFFFFFFFFFFFFLL)
#define SMALLEST_FINITE_DOUBLE  (0xFFEFFFFFFFFFFFFFLL)
#define POSITIVE_INFINITY       (0x7FF0000000000000LL)

/* The followings masks can be used to get the sign, exponent,
 * and significand of a double precision floating point value. */
#define SIGN_MASK               (0x8000000000000000LL)
#define EXPONENT_MASK           (0x7FF0000000000000LL)
#define SIGNIFICAND_MASK        (0x000FFFFFFFFFFFFFLL)

void xl__ieee (signal, info, context)
  int signal;
  siginfo_t *info;
  void *context;
{
  /* The fpstat variable is used as a local copy of the FPSCR
   * value at the point where the signal occurred. */
  fptrap_t fpstat;

  /* The 'instr' variable will be used to examine the code at the
   * point where the signal was generated. */
  unsigned int *instr;

  /* These variables are used when the instruction that caused the
   * exception is decoded. */
  int opcd, xo, frt, fra, frb, frc;  /* instruction fields */
  int num_sources, source[3], src;

  /* The 'ifpr' variable will be used to examine the saved floating-
   * point registers in the sigcontext structure as binary values.
   * The 'sign', 'signif' and 'expon' variables are used when it is
   * necessary to extract and manipulate the fields of a double
   * precision register value.  The 'G' and 'X' variables are used
   * for denormalization of underflow results. */
  long long *ifpr;
  long long sign, signif;
  long expon, G, X;

#if defined(__APPLE__) && defined(__64BIT__)
  ucontext64_t *contextData = (ucontext64_t *) context;
#else
  ucontext_t *contextData = (ucontext_t *) context;
#endif

  /* The 'fpscr' variable will be used to access the Floating
   * Point Status and Control Register. */
  unsigned int *fpscr;

  /* The isNaNQTrap flag, when set, indicated whether a -qflttrap=nanq
   * trap was detected. */
  int isNaNQTrap = 0;

  /* Executable code for the signal handler begins here. */

  /* Call the routine xl__sigdump, from libxlf90.a, to print
   * information about the signal that was received and
   * a traceback starting from the point where the signal
   * was generated. */
  extern void xl__sigdump (int, siginfo_t *, void *);
  xl__sigdump (signal, info, context);
#if defined(__linux__) && \
    (!defined(REDHAT) || defined(__64BIT__) || defined(CELL_PU))
  /* Point to the lower 4-bytes of the FPSCR. */
  fpscr = ((unsigned int *)
          &(contextData->uc_mcontext.regs->gpr[PT_FPSCR])) + 1;
#elif defined(__linux__) && defined(REDHAT) && !defined(__64BIT__)
  /* Point to the lower 4-bytes of the FPSCR. */
  fpscr = ((unsigned int *)
          &(contextData->uc_mcontext.uc_regs->fpregs.fpscr)) + 1;
#elif defined(__APPLE__)
#if defined(__64BIT__)
  fpscr = (unsigned int *) &(contextData->uc_mcontext64->fs.fpscr);
#else
  fpscr = (unsigned int *) &(contextData->uc_mcontext->fs.fpscr);
#endif
#else /* AIX */
  fpscr = &(contextData->uc_mcontext.jmp_context.fpscr);
#endif /* __linux__ */

  /* Prepare for a return to the instruction following the point
   * where the signal was generated.
   * First, make local copies of the floating-point status and
   * control register and the iar register from the point where
   * the exception occurred. */
  fpstat = *fpscr;
#if defined(__linux__) && \
    (!defined(REDHAT) || defined(__64BIT__) || defined(CELL_PU))
  instr = (unsigned int *) contextData->uc_mcontext.regs->nip;
#elif defined(__linux__) && defined(REDHAT) && !defined(__64BIT__)
  instr = (unsigned int *) contextData->uc_mcontext.uc_regs->gregs[PT_NIP];
#elif defined(__APPLE__)
#if defined(__64BIT__)
  instr = (unsigned int *) contextData->uc_mcontext64->ss.srr0;
#else
  instr = (unsigned int *) contextData->uc_mcontext->ss.srr0;
#endif
#else /* AIX */
  instr = (unsigned int *) contextData->uc_mcontext.jmp_context.iar;
#endif /* __linux__ */

  /* Next, clear all exception bits in the saved floating-point
   * status and control register; this will prevent future
   * signals due to floating-point exceptions that occurred
   * before this signal. */
  *fpscr &= ~FP_ALL_XCP;

  /* When a signal is received, the saved iar points at the failing
   * instruction; increment the iar so that the return from the
   * signal handler does not immediately fail again. */
#if defined(__linux__) && \
    (!defined(REDHAT) || defined(__64BIT__) || defined(CELL_PU))
  contextData->uc_mcontext.regs->nip += 4;
#elif defined(__linux__) && defined(REDHAT) && !defined(__64BIT__)
  contextData->uc_mcontext.uc_regs->gregs[PT_NIP] += 4;
#elif defined(__APPLE__)
#if defined(__64BIT__)
  contextData->uc_mcontext64->ss.srr0 += 4;
#else
  contextData->uc_mcontext->ss.srr0 += 4;
#endif
#else /* AIX */
  contextData->uc_mcontext.jmp_context.iar += 4;
#endif /* __linux__ */

  if (signal == SIGTRAP) {
    /* Check that the trap is of the type used for the FLTTRAP
     * option.  If it is not, terminate execution (other trap
     * causes, such as an integer division by zero or a subscript
     * out of range, are best treated as fatal errors). */
    if (*instr != FLTTRAP_INST) {
      fputs("xl__ieee: not a FLTTRAP exception\n", stderr);
      exit(1);
    }
    /* Examine the code sequence leading to the trap to ensure it
     * was generated by the FLTTRAP option.
     * First, ensure the instruction preceding the trap is the expected
     * conditional branch generated by the FLTTRAP option. */
    instr--;
    if (*instr == FLTTRAP_BRANCH) {
      /* The instruction before the branch should be a floating-point
       * operation with record mode. */
      instr--;
      /* Make sure the record bit is set. */
      if ((*instr & 1) != 1) {
        fputs("xl__ieee: invalid FLTTRAP code sequence\n", stderr);
        exit(1);
      }
    }
    else if (*instr == FLTTRAP_NANQ_BRANCH) {
      isNaNQTrap = 1;

      /* The instruction before the branch should be a floating-point
       * compare unordered (fcmpu) instruction.
       * The instruction before that is the real instruction that caused
       * the exception. */
      instr -= 2;

      /* If both nanq and invalid operation trapping are on, we will now
       * be pointing at a trap instruction.
       * e.g. fmuls, bc, tw (for inv), fcmpu, bc, tw (for nanq)
       * So, subtract 2 more instructions in this case. */
      if (*instr == FLTTRAP_INST && *(instr-1) == FLTTRAP_BRANCH)
      {
        instr -= 2;
      }
    }
    else {
      fputs("xl__ieee: invalid FLTTRAP code sequence\n", stderr);
      exit(1);
    }

  }
  else if (signal != SIGFPE) {
    fputs("xl__ieee: not a SIGTRAP or SIGFPE signal\n", stderr);
    exit(1);
  }

  /* Floating-point operations have an opcode field value of 63
   * (double precision) or 59 (single precision). */
  opcd = *instr >> 26;
  if ((opcd != 63) && (opcd != 59)) {

    switch (opcd)
    {
      case 48: /* lfs */
      case 49: /* lfsu */
      case 31: /* lfsux, lfdux, lfqux, lfsx, lfdx, lfqx */
      case 50: /* lfd */
      case 51: /* lfdu */
      case 56: /* lfq */
      case 57: /* lfqu */
        /* -qflttrap=nanq
         * The -qflttrap=nanq option stops on loads of NaN values in addition
         * to IEEE floating point operations.
         * If we got here, then the instruction is a load instruction and we
         * should leave the handler. */
        if (isNaNQTrap)
          return;
      default:
        fputs("xl__ieee: exception point cannot be determined\n", stderr);
        exit(1);
    }
  }

  /* Decode the XO field of the instruction, assuming it is
   * A-form. */
  xo = (*instr >> 1) & 0x1F;

  /* Check to see if the instruction found was a move register (fmr). */
  if (xo == 8) {  /* move register */
    /* A move register is used to detect a floating-point
     * exception that occurred in a previous instruction.
     * This is done after calls to external routines (to see if
     * they returned with any exception bits set), after
     * floating-point compares (which do not have a record mode
     * form), and when the 'imprecise' suboption of FLTTRAP is used.
     * For compares, the compare operation will be just before the
     * move register. */
    instr--;
    opcd = *instr >> 26;
    xo = (*instr >> 1) & 0x1F;
    if ((opcd != 63) || (xo != 0)) {  /* not a compare */
      /* For calls, any external routine that generated an exception
       * but did not generate a SIGTRAP must be a library routine
       * or be in user code that was not compiled with FLTTRAP.  For
       * those cases, terminate execution because the original failing
       * operation cannot be found.
       * Similarly, if the 'imprecise' FLTTRAP mode was used, the
       * failing operation cannot be found. */
      fputs("xl__ieee: exception point cannot be determined\n", stderr);
      exit(1);
    }
    /* else: found the floating point compare that caused the exception */
  }

  /* Decode the rest of the instruction. */
  frt = (*instr >> 21) & 0x1F;
  fra = (*instr >> 16) & 0x1F;
  frb = (*instr >> 11) & 0x1F;
  frc = (*instr >> 6) & 0x1F;

  /* Use the opcode and register information to analyze the
   * failing instruction. */
  switch(xo) {
    case(21):  /* add operation */
      num_sources = 2;
      source[0] = fra;
      source[1] = frb;
      break;
    case(20):  /* subtract operation */
      num_sources = 2;
      source[0] = fra;
      source[1] = frb;
      break;
    case(25):  /* multiply operation */
      num_sources = 2;
      source[0] = fra;
      source[1] = frc;
      break;
    case(18):  /* divide operation */
      num_sources = 2;
      source[0] = fra;
      source[1] = frb;
      break;
    case(29):  /* multiply-add operation */
      num_sources = 3;
      source[0] = fra;
      source[1] = frb;
      source[2] = frc;
      break;
    case(28):  /* multiply-subtract operation */
      num_sources = 3;
      source[0] = fra;
      source[1] = frb;
      source[2] = frc;
      break;
    case(31):  /* negative multiply-add operation */
      num_sources = 3;
      source[0] = fra;
      source[1] = frb;
      source[2] = frc;
      break;
    case(30):  /* negative multiply-subtract operation */
      num_sources = 3;
      source[0] = fra;
      source[1] = frb;
      source[2] = frc;
      break;
    case(12):  /* round to single-precision operation */
      num_sources = 1;
      source[0] = frb;
      break;
    case(14):  /* convert to/from integer */
      num_sources = 1;
      source[0] = frb;
      break;
    case(15):  /* convert to integer with round toward zero */
      num_sources = 1;
      source[0] = frb;
      break;
    case(0):  /* compare operation */
      /* Compares can only generate invalid operation exceptions, and
       * do so when one of the operands of the compare is a NaN.
       * No fixup is needed (the condition register result is set to
       * indicate unordered), so just continue execution. */
      return;
    case(22):  /* square root */
      num_sources = 1;
      source[1] = frb;
      break;
    case(24):  /* reciprocal estimate */
      num_sources = 1;
      source[0] = frb;
      break;
    case(26):  /* reciprocal square root estimate */
      num_sources = 1;
      source[0] = frb;
      break;
    default:
      /* This could be an exception caused by explicitly setting
       * bits in the floating-point status and control register. */
      fputs("xl__ieee: exception point cannot be determined\n", stderr);
      exit(1);
  }

  /* Clear the FPRF field of the FPSCR in preparation for setting
   * it to indicate the class of the result returned below. */
  *fpscr &= ~FP_FPRF_MASK;

  /* Examine the floating point status and control register for
   * enabled exceptions, and set the result register to the
   * IEEE default result. */
#if defined(__linux__) && \
    (!defined(REDHAT) || defined(__64BIT__) || defined(CELL_PU))
  ifpr = (long long *) &(contextData->uc_mcontext.regs->gpr[PT_FPR0]);
#elif defined(__linux__) && defined(REDHAT) && !defined(__64BIT__)
  ifpr = (long long *) contextData->uc_mcontext.uc_regs->fpregs.fpregs;
#elif defined(__APPLE__)
#if defined(__64BIT__)
  ifpr = (long long *) contextData->uc_mcontext64->fs.fpregs;
#else
  ifpr = (long long *) contextData->uc_mcontext->fs.fpregs;
#endif
#else /* AIX */
  ifpr = (long long *) contextData->uc_mcontext.jmp_context.fpr;
#endif /* __linux__ */

  if ((fpstat & (TRP_INVALID|FP_INVALID)) == (TRP_INVALID|FP_INVALID)) {
    /* The default result for a disabled invalid operation depends
     * on the instruction. */
    if ((xo == 14) || (xo == 15)) {  /* convert to integer */
      /* If the source operand is negative or a NaN, the result
       * is the most negative integer, else the result is the
       * the most positive integer. */
      if (((ifpr[frb] & SIGN_MASK) != 0) ||
          (((ifpr[frb] & EXPONENT_MASK) == EXPONENT_MASK) &&
           ((ifpr[frb] & SIGNIFICAND_MASK) != 0))) {
        if (frc == 0)  /* 32-bit result */
          ifpr[frt] = 0x0000000080000000LL;
        else
          ifpr[frt] = 0x8000000000000000LL;
      }
      else {
        if (frc == 0)  /* 32-bit result */
          ifpr[frt] = 0x000000007FFFFFFFLL;
        else
          ifpr[frt] = 0x7FFFFFFFFFFFFFFFLL;
      }
      /* The FPRF field of the FPSCR is undefined for these conversions. */
    }
    else {  /* not a convert to integer operation */
      /* For all other cases, the result is a quiet NaN.  If one
       * of the source operands is a signaling NaN, the result
       * should be a quiet version of that NaN; otherwise, the
       * result is a generic quiet NaN. */
      if (fpstat & FP_INV_SNAN) {
        for (src=0;src<num_sources;src++) {
          if (((ifpr[source[src]] & 0x7FF8000000000000LL) ==
               EXPONENT_MASK) &&
              ((ifpr[source[src]] & SIGNIFICAND_MASK) != 0)) {
            ifpr[frt] = ifpr[source[src]] | 0x0008000000000000LL;
            break;
          }
        }
        if (xo == 12) {  /* round to single-precision operation */
          /* Truncate the fraction to produce a valid single-precision
           * result. */
          ifpr[frt] = ifpr[frt] & 0xFFFFFFFFE0000000LL;
        }
      }
      else
        ifpr[frt] = 0x7ff8000000000000LL;  /* generic quiet nan */
      /* Set the FPRF field of the FPSCR to indicate a quiet NaN result. */
      *fpscr |= FP_FPRF_QNAN;
    }
  }  /* End of invalid operation processing code. */

  else if ((fpstat & (TRP_DIV_BY_ZERO|FP_DIV_BY_ZERO)) ==
           (TRP_DIV_BY_ZERO|FP_DIV_BY_ZERO)) {
    /* The default result is infinity, where the sign is determined
     * by the XOR of the signs of the operands. */
    ifpr[frt] = POSITIVE_INFINITY |
        ((ifpr[fra] ^ ifpr[frb]) & SIGN_MASK);
    /* Set the FPRF field of the FPSCR to indicate an infinity result. */
    if ((ifpr[frt] & SIGN_MASK) == 0)
      *fpscr |= FP_FPRF_PINF;
    else
      *fpscr |= FP_FPRF_NINF;
  }  /* End of divide by zero processing code. */

  else if ((fpstat & (TRP_OVERFLOW|FP_OVERFLOW)) ==
           (TRP_OVERFLOW|FP_OVERFLOW)) {
    /* The result register in an overflow contains a rounded
     * normalized number, but 1536 (double precision) or 192
     * (single precision) has been subtracted from the exponent.
     * The default result depends on the rounding mode and sign
     * of the result. */
    sign = ifpr[frt] & SIGN_MASK;
    switch (fpstat & 3) {
      case(0):  /* round to nearest */
        /* Set the result register to +/- infinity. */
        ifpr[frt] = POSITIVE_INFINITY | sign;
        if (sign == 0)
          *fpscr |= FP_FPRF_PINF;
        else
          *fpscr |= FP_FPRF_NINF;
        break;

      case(1):  /* round toward zero */
        /* Set the result register to the largest finite floating
         * point number. */
        if ((opcd == 59) || (xo == 12))  /* single precision */
          ifpr[frt] = LARGEST_FINITE_SINGLE;
        else
          ifpr[frt] = LARGEST_FINITE_DOUBLE;
        ifpr[frt] = ifpr[frt] | sign;
        if (sign == 0)
          *fpscr |= FP_FPRF_PNORM;
        else
          *fpscr |= FP_FPRF_NNORM;
        break;

      case(2):  /* round toward +infinity */
        /* If the number is positive, set the result register
           to +infinity.
           If the number is negative, set the result register to the
           most negative finite floating point number. */
        if (sign == 0) {
          ifpr[frt] = POSITIVE_INFINITY;
          *fpscr |= FP_FPRF_PINF;
        }
        else {
          if ((opcd == 59) || (xo == 12))  /* single precision */
            ifpr[frt] = SMALLEST_FINITE_SINGLE;
          else
            ifpr[frt] = SMALLEST_FINITE_DOUBLE;

          *fpscr |= FP_FPRF_NNORM;
        }
        break;

      case(3):  /* round toward -infinity */
        /* If the number is negative, set the result register
           to -infinity.
           If the number is positive, set the result register
           to the most positive finite floating point number. */
        if (sign == 0) {
          if ((opcd == 59) || (xo == 12))  /* single precision */
            ifpr[frt] = LARGEST_FINITE_SINGLE;
          else
            ifpr[frt] = LARGEST_FINITE_DOUBLE;

          *fpscr |= FP_FPRF_PNORM;
        }
        else {
          ifpr[frt] = 0xFFF0000000000000LL;  /* -infinity */
          *fpscr |= FP_FPRF_NINF;
        }
        break;
    }
    /* The FPSCR FR bit is undefined, and the FI bit should be set. */
    *fpscr |= FP_FI;
  }  /* End of overflow processing code. */

  else if ((fpstat & (TRP_UNDERFLOW|FP_UNDERFLOW)) ==
           (TRP_UNDERFLOW|FP_UNDERFLOW)) {
    /* The result register in an underflow contains a rounded
     * normalized number, but 1536 (double precision) or 192
     * (single precision) has been added to the exponent.
     * The default result depends on the rounding mode and sign
     * of the result. */
    sign = ifpr[frt] & SIGN_MASK;
    expon = (ifpr[frt] >> 52) & 0x7FF;
    signif = (ifpr[frt] & SIGNIFICAND_MASK) | 0x0010000000000000LL;
    if ((opcd == 59) || (xo == 12)) {  /* single precision */
      expon = expon - 192;
      /* Shift the double precision significand 29 bits right to get
       * the short precision significand. */
      signif = signif >> 29;
    }
    else
      expon = expon - 1536;
    /* The 'expon' value is now a negative number, and represents the
     * right shift amount that must be applied to the significand
     * (that is, by shifting the significand and incrementing the
     * exponent until the exponent has value one, we can derive the
     * fraction for the default result).
     * The FR and FI bits in the FPSCR allow us to 'unround' the
     * result produced.  If set, the FI bit indicates that some
     * type of rounding was applied, depending on the rounding
     * mode and the G, R and X bits.  If the FR bit is set, the
     * significand was incremented as part of the rounding.
     * For our purposes, we will need to maintain a version of the
     * guard (G) bit, and a sticky (X) bit.  The round (R) bit does
     * not need to be explicitly maintained, so our X bit will
     * serve as a logical OR of the R and X bits specified by an
     * IEEE execution model.
     * Initially, assume no rounding. */
    G = 0;
    X = 0;
    if (fpstat & FP_FI) {
      /* Some rounding was applied, depending on the rounding mode.
       * For our purposes, it is sufficient to assume that the X
       * bit should be set. */
      X = 1;
      /* If the significand was incremented as part of rounding, then
       * decrement it now and set the G bit, which is again sufficient
       * for our purposes. */
      if (fpstat & FP_FR) {
        signif = signif - 1;
        G = 1;
        /* Clear the FR bit in the saved FPSCR; it will be set below
         * based on the rounding of the default result. */
        *fpscr &= ~FP_FR;
      }
    }
    /* Now shift the significand right to denormalize the result.
     * For efficiency, stop shifting if both the significand and
     * the guard bit are zero. */
    while ((expon <= 0) && ((signif != 0) || (G != 0))) {
      X |= G;
      G = signif & 1;
      signif = signif >> 1;
      expon = expon + 1;
    }
    if ((G != 0) || (X != 0)) {
      /* Result is inexact.  Make sure FI is set in the FPSCR, and
       * use to rounding mode to complete the rounding of the
       * denormalized result. */
      *fpscr |= FP_FI;
      switch (fpstat & 3) {
        case(0):  /* round to nearest */
          if (G && (X || (signif & 1))) {
            signif = signif + 1;
            *fpscr |= FP_FR;
          }
          break;
        case(1):  /* round toward zero */
          break;
        case(2):  /* round toward +infinity */
          if (sign == 0) {
            signif = signif + 1;
            *fpscr |= FP_FR;
          }
          break;
        case(3):  /* round toward -infinity */
          if (sign != 0) {
            signif = signif + 1;
            *fpscr |= FP_FR;
          }
          break;
      }
    }
    if ((opcd == 59) || (xo == 12)) {  /* single precision */
      /* Shift the short precision significand 29 bits left to get
       * a double precision significand for the result register. */
      signif = signif << 29;
    }
    /* The final result is the original sign, a zero exponent, and
     * the derived significand. */
    ifpr[frt] = sign | signif;
    if (sign == 0) {
      if (signif == 0)
        /* result is positive zero */
        *fpscr |= FP_FPRF_PZERO;
      else
        /* result is a positive denormal */
        *fpscr |= FP_FPRF_PDEN;
    }
    else {
      if (signif == 0)
        /* result is negative zero */
        *fpscr |= FP_FPRF_NZERO;
      else
        /* result is a negative denormal */
        *fpscr |= FP_FPRF_NDEN;
    }
  }  /* End of underflow processing code. */

  else if ((fpstat & (TRP_INEXACT|FP_INEXACT)) ==
          (TRP_INEXACT|FP_INEXACT)) {
    /* No action necessary, just ignore this. */
  }
}


/* ---------------------------------------------------------------------
! * The entry point xl__ieee_ is provided for programs that use
! * xl__ieee but are compiled with -qextname.
!-------------------------------------------------------------------- */

void xl__ieee_ (signal, info, context)
  int signal;
  siginfo_t *info;
  void *context;
{
    xl__ieee (signal, info, context);
}
